@@ -6,22 +6,25 @@ from time import sleep
 from typing import List, Optional, AnyStr, TypeVar, TextIO, Tuple, Callable, Dict
 from urllib.parse import urlparse, urljoin
 
+from bs4 import BeautifulSoup
 from loguru import logger
 from requests import Response, Session, Request, PreparedRequest, codes
 # noinspection PyUnresolvedReferences,PyProtectedMember
 from requests._internal_utils import to_native_string
 from requests.adapters import BaseAdapter
 from requests.cookies import extract_cookies_to_jar, merge_cookies, cookiejar_from_dict
-from requests.exceptions import ChunkedEncodingError, ContentDecodingError, TooManyRedirects
+from requests.exceptions import ChunkedEncodingError, ContentDecodingError, \
+    TooManyRedirects, RequestException
 from requests.sessions import merge_setting, merge_hooks
 from requests.structures import CaseInsensitiveDict
 from requests.utils import requote_uri, rewind_body, get_netrc_auth
 from urllib3.util.url import parse_url, Url
 
 from spoofbot.adapter import FileCache
-from spoofbot.operating_system import Windows
+from spoofbot.operating_system import Windows, random_os, OS
 from spoofbot.tag import MimeTypeTag, LanguageTag
-from spoofbot.util import ReferrerPolicy, are_same_origin, are_same_site, sort_dict, TimelessRequestsCookieJar
+from spoofbot.util import ReferrerPolicy, are_same_origin, are_same_site, sort_dict, \
+    TimelessRequestsCookieJar, random_version, get_firefox_versions, get_chrome_versions
 from spoofbot.util.log import log_request, log_response
 
 
@@ -70,7 +73,8 @@ class User(Enum):
 
 
 DictOrBytes = TypeVar('DictOrBytes', dict, bytes)
-DictOrTupleListOrBytesOrFileLike = TypeVar('DictOrTupleListOrBytesOrFileLike', dict, List[tuple], bytes, TextIO)
+DictOrTupleListOrBytesOrFileLike = TypeVar('DictOrTupleListOrBytesOrFileLike', dict,
+                                           List[tuple], bytes, TextIO)
 DictOrCookieJar = TypeVar('DictOrCookieJar', dict, CookieJar)
 StrOrFileLike = TypeVar('StrOrFileLike', str, TextIO)
 AuthTupleOrCallable = TypeVar('AuthTupleOrCallable', Tuple[str, str], Callable)
@@ -106,6 +110,9 @@ class Browser(Session):
 
     def __init__(self):
         super(Browser, self).__init__()
+        self._name = 'Spoofbot'
+        from spoofbot import __version__
+        self._version = __version__
         self._user_agent = ''
         self._accept = []
         self._accept_language = []
@@ -128,6 +135,16 @@ class Browser(Session):
         # noinspection PyTypeChecker
         self._adapter = self.get_adapter('https://')
 
+    @property
+    def name(self) -> str:
+        """Name of the browser"""
+        return self._name
+
+    @property
+    def version(self) -> str:
+        """Version of the browser"""
+        return self._version
+
     @property
     def adapter(self) -> BaseAdapter:
         """Gets the adapter for the HTTP/HTTPS requests
@@ -146,6 +163,7 @@ class Browser(Session):
         """
         self._adapter = adapter
         self.mount('https://', adapter)
+        # noinspection HttpUrlsUsage
         self.mount('http://', adapter)
 
     @property
@@ -277,6 +295,15 @@ class Browser(Session):
         """
         raise NotImplementedError
 
+    @staticmethod
+    def create_random_user_agent() -> str:
+        """Creates seemingly random user agent string
+
+        :returns: A random user agent string.
+        :rtype: str
+        """
+        raise NotImplementedError
+
     # noinspection DuplicatedCode
     def _get_referer(self, url: Url) -> Optional[str]:
         if self._last_navigate is None:
@@ -291,8 +318,8 @@ class Browser(Session):
         if not are_same_origin(nav_url, url) or method not in ['GET', 'HEAD']:
             return self._referrer_policy.get_origin(nav_url, url)
 
-    # noinspection PyMethodMayBeStatic
-    def _get_host(self, url: Url) -> str:
+    @staticmethod
+    def _get_host(url: Url) -> str:
         if url.port:
             return f"{url.hostname}:{url.port}"
         return url.hostname
@@ -334,8 +361,8 @@ class Browser(Session):
         if url.scheme == 'https' and self._te != '':
             return self._te
 
-    # noinspection PyMethodMayBeStatic
-    def _get_sec_fetch_dest(self, dest: Destination) -> str:
+    @staticmethod
+    def _get_sec_fetch_dest(dest: Destination) -> str:
         # https://www.w3.org/TR/fetch-metadata/#sec-fetch-dest-header
         # noinspection SpellCheckingInspection
         if dest is None:
@@ -373,8 +400,9 @@ class Browser(Session):
         # noinspection PyTypeChecker
         return site.value
 
-    def navigate(self, url: str, **kwargs) -> Response:
-        """Sends a GET request to the url and sets it into the Referer header in subsequent requests
+    def navigate(self, url: str, **kwargs) -> list[Response]:
+        """Sends a GET request to the url and sets it into the Referer header in
+        subsequent requests
 
         :param url: The url the browser is supposed to connect to
         :param kwargs: Additional arguments to forward to the requests module
@@ -384,13 +412,87 @@ class Browser(Session):
         kwargs.setdefault('user_activation', True)
         response = self.get(url, **kwargs)
         self._last_navigate = response
-        return response
+        return self._request_attachments(response)
+
+    def _request_attachments(self, response: Response) -> list[Response]:
+        response.raise_for_status()
+        responses = [response]
+        bs = BeautifulSoup(response.content, features='html.parser')
+        url = parse_url(response.url)
+        links = self._gather_valid_links(bs, url) + \
+                self._gather_valid_scripts(bs, url) + \
+                self._gather_valid_imgs(bs, url)
+        for link in links:
+            logger.debug(f"Fetching {link.url}")
+            try:
+                resp = self.get(link.url)
+                responses.append(resp)
+            except RequestException:
+                pass
+        return responses
+
+    @staticmethod
+    def _gather_valid_links(bs: BeautifulSoup, origin: Url) -> list[Url]:
+        links = []
+        for link in bs.find_all('link'):
+            ignore = False
+            for rel in link.get('rel', None):
+                # https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes/rel
+                if rel in {'dns-prefetch', 'preconnect'}:
+                    # DNS resolve and preemptively connecting is useless to us
+                    ignore |= True
+                elif rel in {'canonical'}:
+                    ignore |= True
+                elif rel in {'manifest', 'mask-icon'}:
+                    # non-standard and unsupported
+                    ignore |= True
+            if ignore:
+                continue
+            href: str = link.get('href', None)
+            if href is None:
+                continue
+            if href.startswith('/'):
+                href = f"{origin.scheme}://{origin.hostname}{href}"
+            links.append(parse_url(href))
+        return links
+
+    @staticmethod
+    def _gather_valid_scripts(bs: BeautifulSoup, origin: Url) -> list[Url]:
+        scripts = []
+        for script in bs.find_all('script'):
+            if 'src' not in script.attrs:
+                continue
+            src: str = script.get('src', None)
+            if src is None:
+                continue
+            if src.startswith('/'):
+                src = f"{origin.scheme}://{origin.hostname}{src}"
+            scripts.append(parse_url(src))
+        return scripts
+
+    @staticmethod
+    def _gather_valid_imgs(bs: BeautifulSoup, origin: Url) -> list[Url]:
+        scripts = []
+        for script in bs.find_all('img'):
+            if 'src' not in script.attrs:
+                continue
+            src: str = script.get('src', None)
+            if src is None:
+                continue
+            if src.startswith('/'):
+                src = f"{origin.scheme}://{origin.hostname}{src}"
+            scripts.append(parse_url(src))
+        return scripts
 
     def request(self, method: AnyStr, url: AnyStr, params: DictOrBytes = None,
-                data: DictOrTupleListOrBytesOrFileLike = None, headers: dict = None, cookies: DictOrCookieJar = None,
-                files: StrOrFileLike = None, auth: AuthTupleOrCallable = None, timeout: FloatOrTuple = None,
-                allow_redirects=True, proxies: Dict[str, str] = None, hooks: Dict[str, Callable] = None,
-                stream: bool = None, verify: StrOrBool = None, cert: StrOrStrTuple = None,
+                data: DictOrTupleListOrBytesOrFileLike = None, headers: dict = None,
+                cookies: DictOrCookieJar = None,
+                files: StrOrFileLike = None, auth: AuthTupleOrCallable = None,
+                timeout: FloatOrTuple = None,
+                allow_redirects=True, proxies: Dict[str, str] = None,
+                hooks: Dict[str, Callable] = None,
+                stream: bool = None, verify: StrOrBool = None,
+                cert: StrOrStrTuple = None,
                 json: str = None, user_activation: bool = False) -> Response:
         """Constructs a :class:`Request <Request>`, prepares it and sends it.
         Returns :class:`Response <Response>` object.
@@ -425,15 +527,16 @@ class Browser(Session):
             hostname to the URL of the proxy.
         :param stream: (optional) whether to immediately download the response
             content. Defaults to ``False``.
-        :param verify: (optional) Either a boolean, in which case it controls whether we verify
-            the server's TLS certificate, or a string, in which case it must be a path
-            to a CA bundle to use. Defaults to ``True``.
+        :param verify: (optional) Either a boolean, in which case it controls whether we
+            verify the server's TLS certificate, or a string, in which case it must
+            be a path to a CA bundle to use. Defaults to ``True``.
         :param cert: (optional) if String, path to ssl client cert file (.pem).
             If Tuple, ('cert', 'key') pair.
         :rtype: requests.Response
         """
         cookies = cookies if cookies is not None else self.cookies
-        self.headers = self._get_default_headers(method, parse_url(url), user_activation)
+        self.headers = self._get_default_headers(method, parse_url(url),
+                                                 user_activation)
         self.headers.update(headers if headers else {})
 
         # Create the Request.
@@ -453,7 +556,8 @@ class Browser(Session):
 
         log_request(prep)
 
-        prep.headers = CaseInsensitiveDict(sort_dict(dict(prep.headers), self._header_precedence))
+        prep.headers = CaseInsensitiveDict(
+            sort_dict(dict(prep.headers), self._header_precedence))
 
         proxies = proxies or {}
 
@@ -499,7 +603,8 @@ class Browser(Session):
         # Merge with session cookies
         cookie_jar = self.cookies.__new__(self.cookies.__class__)
         cookie_jar.__init__()
-        if isinstance(cookie_jar, TimelessRequestsCookieJar) and isinstance(self.cookies, TimelessRequestsCookieJar):
+        if isinstance(cookie_jar, TimelessRequestsCookieJar) and isinstance(
+                self.cookies, TimelessRequestsCookieJar):
             cookie_jar.mock_date = self.cookies.mock_date
         merged_cookies = merge_cookies(
             merge_cookies(cookie_jar, self.cookies), cookies)
@@ -516,18 +621,21 @@ class Browser(Session):
             files=request.files,
             data=request.data,
             json=request.json,
-            headers=merge_setting(request.headers, self.headers, dict_class=CaseInsensitiveDict),
+            headers=merge_setting(request.headers, self.headers,
+                                  dict_class=CaseInsensitiveDict),
             params=merge_setting(request.params, self.params),
             auth=merge_setting(auth, self.auth),
             cookies=merged_cookies,
             hooks=merge_hooks(request.hooks, self.hooks),
         )
-        p.headers = CaseInsensitiveDict(sort_dict(dict(p.headers), self._header_precedence))
+        p.headers = CaseInsensitiveDict(
+            sort_dict(dict(p.headers), self._header_precedence))
         return p
 
     # noinspection PyUnresolvedReferences
     def resolve_redirects(self, resp, req, stream=False, timeout=None,
-                          verify=True, cert=None, proxies=None, yield_requests=False, **adapter_kwargs):
+                          verify=True, cert=None, proxies=None, yield_requests=False,
+                          **adapter_kwargs):
         """Receives a Response. Returns a generator of Responses or Requests."""
 
         hist = []  # keep track of history
@@ -550,7 +658,8 @@ class Browser(Session):
                 resp.raw.read(decode_content=False)
 
             if len(resp.history) >= self.max_redirects:
-                raise TooManyRedirects('Exceeded %s redirects.' % self.max_redirects, response=resp)
+                raise TooManyRedirects('Exceeded %s redirects.' % self.max_redirects,
+                                       response=resp)
 
             # Release the connection back into the pool.
             resp.close()
@@ -583,7 +692,8 @@ class Browser(Session):
             self.rebuild_method(prepared_request, resp)
 
             # https://github.com/requests/requests/issues/1084
-            if resp.status_code not in (codes.temporary_redirect, codes.permanent_redirect):
+            if resp.status_code not in (
+                    codes.temporary_redirect, codes.permanent_redirect):
                 # https://github.com/requests/requests/issues/3490
                 purged_headers = ('Content-Length', 'Content-Type', 'Transfer-Encoding')
                 for header in purged_headers:
@@ -652,7 +762,8 @@ class Browser(Session):
                 rewind_body(prepared_request)
 
             # Override the original request.
-            prepared_request.headers = dict(sort_dict(prepared_request.headers, self._header_precedence))
+            prepared_request.headers = dict(
+                sort_dict(prepared_request.headers, self._header_precedence))
             req = prepared_request
             log_request(req)
 
@@ -678,16 +789,19 @@ class Browser(Session):
                 self._last_navigate = resp
                 yield resp
 
-    def _get_default_headers(self, method: str, url: Url, user_activation: bool) -> CaseInsensitiveDict:
-        """Provides the default headers the browser should send when connecting to an endpoint
+    def _get_default_headers(self, method: str, url: Url,
+                             user_activation: bool) -> CaseInsensitiveDict:
+        """Provides the default headers the browser should send when connecting to an
+        endpoint
 
-        The method tries to guess the mimetype and encoding to fill the Accept and Accept-Encoding headers
+        The method tries to guess the mimetype and encoding to fill the Accept and
+        Accept-Encoding headers
         :param method: The method of the HTTP request
         :param url: The url the browser is supposed to connect to
         :returns: A dictionary form of the default headers.
         :rtype: OrderedHeaders
         """
-        return CaseInsensitiveDict({
+        return CaseInsensitiveDict(dict(filter(lambda kvp: kvp[1] != '', {
             'Host': self._get_host(url),
             'User-Agent': self._get_user_agent(),
             'Accept': self._get_accept(url),
@@ -699,7 +813,7 @@ class Browser(Session):
             'DNT': self._get_dnt(),
             'Upgrade-Insecure-Requests': self._get_upgrade_insecure_requests(),
             'TE': self._get_te(url),
-        })
+        }.items())))
 
     def await_timeout(self, url: Url = None):
         """Waits until the request timeout expires.
@@ -712,7 +826,8 @@ class Browser(Session):
         time_passed = datetime.now() - self._last_request_timestamp
         if time_passed < self._request_timeout:
             adapter = self.adapter
-            if url is not None and isinstance(adapter, FileCache) and adapter.is_hit(url):
+            if url is not None and isinstance(adapter, FileCache) and adapter.is_hit(
+                    url) and adapter.is_active:
                 logger.debug("Request will be a hit in cache. No need to wait.")
                 return
             time_to_wait = self._request_timeout - time_passed
@@ -737,6 +852,8 @@ class Firefox(Browser):
                  do_not_track=False,
                  upgrade_insecure_requests=True):
         super(Firefox, self).__init__()
+        self._name = 'Firefox'
+        self._version = '.'.join(map(str, ff_version))
         self._user_agent = self.create_user_agent(os, ff_version, build_id)
         self._accept = [
             MimeTypeTag("text", "html"),
@@ -771,7 +888,7 @@ class Firefox(Browser):
         ]
 
     @staticmethod
-    def create_user_agent(os=Windows(), version=FF_NEWEST, build_id=20100101) -> str:
+    def create_user_agent(os: OS = Windows(), version: tuple[int, ...] = FF_NEWEST, build_id: int = 20100101) -> str:
         """Creates a user agent string for Firefox
 
         :param os: The underlying operating system (default :py:class:`Windows`).
@@ -784,6 +901,12 @@ class Firefox(Browser):
                f"Gecko/{build_id} " \
                f"Firefox/{ff_version}"
 
+    @staticmethod
+    def create_random_user_agent() -> str:
+        os = random_os()
+        ff_version = random_version(get_firefox_versions())
+        return Firefox.create_user_agent(os, ff_version)
+
 
 CHROME_NEWEST = (92, 0, 4495, 0)
 WEBKIT_NEWEST = (537, 36)
@@ -797,7 +920,10 @@ class Chrome(Browser):
                  do_not_track=False,
                  upgrade_insecure_requests=True):
         super(Chrome, self).__init__()
-        self._user_agent = self.create_user_agent(os=os, version=chrome_version, webkit_version=webkit_version)
+        self._name = 'Chrome'
+        self._version = '.'.join(map(str, chrome_version))
+        self._user_agent = self.create_user_agent(os=os, version=chrome_version,
+                                                  webkit_version=webkit_version)
         self._accept = [
             MimeTypeTag("text", "html"),
             MimeTypeTag("application", "xhtml+xml"),
@@ -836,11 +962,13 @@ class Chrome(Browser):
         ]
 
     @staticmethod
-    def create_user_agent(os=Windows(), version=CHROME_NEWEST, webkit_version=WEBKIT_NEWEST) -> str:
+    def create_user_agent(os: OS = Windows(), version=CHROME_NEWEST,
+                          webkit_version=WEBKIT_NEWEST) -> str:
         """Creates a user agent string for Firefox
 
         :param os: The underlying operating system (default :py:class:`Windows`).
-        :param version: The version of the underlying webkit (default `(79, 0, 3945, 88)).
+        :param version: The version of the underlying webkit
+                        (default `(79, 0, 3945, 88)).
         :param webkit_version: The version of Chrome (default: (537, 36)).
         :returns: A custom user agent string.
         """
@@ -850,15 +978,22 @@ class Chrome(Browser):
                f"Chrome/{'.'.join(map(str, version))} " \
                f"Safari/{webkit_ver}"
 
-    def navigate(self, url: str, **kwargs) -> Response:
+    @staticmethod
+    def create_random_user_agent() -> str:
+        os = random_os()
+        chrome_version = random_version(get_chrome_versions())
+        return Chrome.create_user_agent(os, chrome_version)
+
+    def navigate(self, url: str, **kwargs) -> List[Response]:
         if parse_url(url).scheme == 'https':
             kwargs.setdefault('headers', {}).setdefault('Sec-Fetch-User', '?1')
             kwargs.setdefault('headers', {}).setdefault('Sec-Fetch-Mode', 'navigate')
-        response = self.get(url, **kwargs)
-        self._last_navigate = response
-        return response
+        responses = super(Chrome, self).navigate(url, **kwargs)
+        self._last_navigate = responses[0]
+        return responses
 
-    def _get_default_headers(self, method: str, url: Url, user_activation: bool) -> CaseInsensitiveDict:
+    def _get_default_headers(self, method: str, url: Url,
+                             user_activation: bool) -> CaseInsensitiveDict:
         adjust_accept_encoding = self._last_navigate is None
         if adjust_accept_encoding:
             self._accept_encoding = ['gzip', 'deflate']
@@ -883,36 +1018,7 @@ class Chrome(Browser):
             if self._last_navigate is None:
                 request.headers['Sec-Fetch-Mode'] = 'navigate'
             else:
-                request.headers['Sec-Fetch-Mode'] = self._get_sec_fetch_mode(request.method, url)
-        request.headers = CaseInsensitiveDict(sort_dict(dict(request.headers), self._header_precedence))
-
-# def build_chromium_user_agent(
-#         os=Linux(),
-#         chromium_version='79.0.3945.92',
-#         webkit_version='607.1.40') -> str:
-#     return f"Mozilla/5.0 ({os}) " \
-#            f"AppleWebKit/{webkit_version} (KHTML, like Gecko) " \
-#            f"Chromium/{chromium_version} " \
-#            f"Safari/{webkit_version}"
-#
-#
-# def build_safari_user_agent(
-#         os=MacOSX(),
-#         safari_version='12.1.2',
-#         webkit_version='607.1.40') -> str:
-#     return f"Mozilla/5.0 ({os}) " \
-#            f"AppleWebKit/{webkit_version} (KHTML, like Gecko) " \
-#            f"Version/{safari_version} " \
-#            f"Safari/{webkit_version}"
-#
-#
-# def build_opera_user_agent(
-#         os=MacOSXVersion.Catalina,
-#         opera_version='65.0.3467.42',
-#         chrome_version='78.0.3904.87',
-#         webkit_version='537.36') -> str:
-#     return f"Mozilla/5.0 ({os.to_comment()}) " \
-#            f"AppleWebKit/{webkit_version} (KHTML, like Gecko) " \
-#            f"Chrome/{chrome_version} " \
-#            f"Safari/{webkit_version} " \
-#            f"OPR/{opera_version}"
+                request.headers['Sec-Fetch-Mode'] = self._get_sec_fetch_mode(
+                    request.method, url)
+        request.headers = CaseInsensitiveDict(
+            sort_dict(dict(request.headers), self._header_precedence))
